# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: executor.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x65xecutor.proto\"j\n\x0f\x45xecutorContext\x12\n\n\x02id\x18\x01 \x01(\t\x12\x15\n\rconnection_id\x18\x02 \x01(\x05\x12\x12\n\nsession_id\x18\x03 \x01(\t\x12\x0f\n\x07session\x18\x04 \x01(\t\x12\x0f\n\x07\x63ontext\x18\x05 \x01(\t\"B\n\x10\x45xecutionContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0b\n\x03sql\x18\x02 \x01(\t\"H\n\x16\x42inaryExecutionContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0b\n\x03sql\x18\x02 \x01(\x0c\"L\n\x17StatementExecuteContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0e\n\x06params\x18\x02 \x01(\t\"E\n\x10\x44\x65\x66\x61ultDBContext\x12!\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x10.ExecutorContext\x12\x0e\n\x06new_db\x18\x02 \x01(\t\"\xbe\x01\n\x10\x45xecutorResponse\x12\x0f\n\x07\x63olumns\x18\x01 \x01(\x0c\x12\x0e\n\x06params\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x13\n\x0bstate_track\x18\x04 \x01(\t\x12\x15\n\rserver_status\x18\x05 \x01(\t\x12\x13\n\x0bis_executed\x18\x06 \x01(\x08\x12\x0f\n\x07session\x18\x07 \x01(\t\x12\x12\n\nerror_code\x18\x08 \x01(\x05\x12\x15\n\rerror_message\x18\t \x01(\t\"@\n\x16\x45xecutorStatusResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t2\xa6\x04\n\x0f\x45xecutorService\x12:\n\x10StatementPrepare\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x41\n\x10StatementExecute\x12\x18.StatementExecuteContext\x1a\x11.ExecutorResponse\"\x00\x12\x36\n\x0cQueryExecute\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x42\n\x12\x42inaryQueryExecute\x12\x17.BinaryExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x39\n\x0f\x45xecuteExternal\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12/\n\x05Parse\x12\x11.ExecutionContext\x1a\x11.ExecutorResponse\"\x00\x12\x32\n\tDoExecute\x12\x10.ExecutorContext\x1a\x11.ExecutorResponse\"\x00\x12\x39\n\x0f\x43hangeDefaultDB\x12\x11.DefaultDBContext\x1a\x11.ExecutorResponse\"\x00\x12=\n\x0e\x44\x65leteExecutor\x12\x10.ExecutorContext\x1a\x17.ExecutorStatusResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'executor_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXECUTORCONTEXT._serialized_start=18
  _EXECUTORCONTEXT._serialized_end=124
  _EXECUTIONCONTEXT._serialized_start=126
  _EXECUTIONCONTEXT._serialized_end=192
  _BINARYEXECUTIONCONTEXT._serialized_start=194
  _BINARYEXECUTIONCONTEXT._serialized_end=266
  _STATEMENTEXECUTECONTEXT._serialized_start=268
  _STATEMENTEXECUTECONTEXT._serialized_end=344
  _DEFAULTDBCONTEXT._serialized_start=346
  _DEFAULTDBCONTEXT._serialized_end=415
  _EXECUTORRESPONSE._serialized_start=418
  _EXECUTORRESPONSE._serialized_end=608
  _EXECUTORSTATUSRESPONSE._serialized_start=610
  _EXECUTORSTATUSRESPONSE._serialized_end=674
  _EXECUTORSERVICE._serialized_start=677
  _EXECUTORSERVICE._serialized_end=1227
# @@protoc_insertion_point(module_scope)
